/*
**	VT100.C
**	By Michael A. Quinlan
**	May 24, 1986
**
** Microsoft C Compiler, Version 3.00
**
**	Terminal emulator for the DEC VT100. Uses ASYNC.ASM for interrupt driven
** communication support. Also uses assembler routines KBD_STAT and KBD_GETC;
** they use the BIOS to read the keyboard in order to bypass PC DOS and its
** infernal Control-C checking.
**
** Functions that are specific to the IBM PC and compatables:
**
**		Everything in ASYNC.ASM. It uses the BIOS for part of the communications
**			port initialization, and makes direct access to the I/O ports. It
**			also intercepts interrupts from the 8250 UART and the timer.
**
**		KBDSTAT.ASM and KBDGETC.ASM use BIOS routines to access the keyboard.
**
**		VIDSETCP, VIDSCROLL, and VIDWRTAC use BIOS routines to write to the
**			video display.
**
**		VIDWRTTTY uses a BIOS routine to make a "bell" sound.
**
** Only supports a minimal subset of the VT100 functions. Supported control
** sequences are:
**
**		ESC [ Pn D			Cursor Backward	
**		ESC [ Pn B			Cursor Down
**		ESC [ Pn C			Cursor Forward
**		ESC [ Pn; Pn H		Cursor Position
**		ESC [ Pn A			Cursor Up
**		ESC [ Ps J			Erase In Display
**		ESC [ Ps K			Erase In Line
**		ESC [ Pn; Pn f		Horizontal and Vertical Position
**		ESC [ Ps;...;Ps m	Select Graphics Rendition. Only bold and normal
**								intensity are supported.
**
** Keyboard translations:
**
**		The Arrow keys on the PC keyboard are translated to the arrow keys
**		on the VT100. F1 thru F4 on the PC keyboard are translated to PF1
**		thru PF4 on the VT100.
**
** Communication Parameters (port, baud, parity, etc.) are hard coded as
** #defines	in this program; you have to re-compile in order to change them.
*/

#define LINT_ARGS
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include "async.h"
#include "kbd.h"

#define FALSE 0
#define TRUE (!FALSE)

/*
** Define communication parameters, as required by the BIOS.
**
** Bits, as defined by the BIOS:
**
**		7			6			5			4			3			2			1			0
**		---- Baud Rate ----			- Parity -		StopBits		- Word Len -
**		000 (0x00) - 110 Baud 		x0 - None		0 - 1 bit	10 - 7 bits
**		001 (0x01) - 150				01 - Odd			1 - 2 bits	11 - 8 bits
**		010 (0x02) - 300				11 - Even
**		011 (0x03) - 600
**		100 (0x04) - 1200
**		101 (0x05) - 2400
**		110 (0x06) - 4800
**		111 (0x07) - 9600
*/
#define COMM_BAUD			0x04		/* 1200 baud */
#define COMM_PARITY		0x00		/* No Parity */
#define COMM_STOPBITS	0x00		/* 1 stop bit */
#define COMM_DATABITS	0x03		/* 8-bit words */
#define COMM_PARM			((COMM_BAUD << 5) | (COMM_PARITY << 3) | \
								(COMM_STOPBITS << 2) | COMM_DATABITS)

#define COMM_PORT			1			/* use COM1 */

/*
** Support Routines for the Video Display
*/

#define VIDEO_INT		0x10		/* BIOS interrupt for Video Display */
#define ATTR_NORM		0x07		/* attribute for normal intensity */
#define ATTR_BOLD		0x0F		/* attribute for bold intensity */
#define ATTR_REV		0x70		/* attribute for reverse video */

int currow, curcol;
int curattr = ATTR_NORM;

void VIDSETCP(row, col)
/* Set Cursor Position */
/* coordinate are converted from one-relative to zero-relative */
int row, col;
{
	union REGS rv;

	if (row < 1) row = 1;
	if (row > 25) row = 25;
	if (col < 1) col = 1;
	if (col > 80) col = 80;

	rv.h.ah = 0x02;			/* Set Cursor Position function */
	rv.h.dh = row - 1;
	rv.h.dl = col - 1;
	rv.h.bh = 0;				/* Video Page */
	int86(VIDEO_INT, &rv, &rv);
	currow = row;
	curcol = col;
}

void VIDSCROLL(numlines, ulrow, ulcol, lrrow, lrcol)
/* Scroll Up an area of the screen */
/* coordinate are converted from one-relative to zero-relative */
int numlines;					/* numlines = 0 means clear screen */
int ulrow, ulcol;				/* upper left corner */
int lrrow, lrcol;				/* lower right corner */
{
	union REGS rv;
	rv.h.ah = 0x06;			/* Scroll Active Page Up function */
	rv.h.al = numlines;
	rv.h.ch = ulrow - 1;
	rv.h.cl = ulcol - 1;
	rv.h.dh = lrrow - 1;
	rv.h.dl = lrcol - 1;
	rv.h.bh = ATTR_NORM;		/* Atribute for new lines */
	int86(VIDEO_INT, &rv, &rv);
}

void VIDWRTAC(a, c, n)
/* Write an attribute/character at the current position */
char a, c;						/* attribute and character */
int n;							/* number of copies to write */
{
	union REGS rv;
	rv.h.ah = 0x09;			/* Write attr/char function */
	rv.h.al = c;				/* character to write */
	rv.h.bh = 0;				/* Display page number */
	rv.h.bl = a;				/* attribute of the character */
	rv.x.cx = n;				/* number of copies to write */
	int86(VIDEO_INT, &rv, &rv);
}

void VIDWRTTTY(c)
/* Write a character in TTY mode */
/* only used to make a "bell" sound */
char c;							/* character to write */
{
	union REGS rv;
	rv.h.ah = 0x0E;			/* Write TTY function */
	rv.h.al = c;				/* character to write */
	rv.h.bl = ATTR_NORM;		/* attribute of the character */
	int86(VIDEO_INT, &rv, &rv);
}

/*
** Routine to cleanup when we finish. Required, otherwise the system will
** crash quite quickly.
**
**  Microsoft C always exits thru the _exit procedure.
*/

static int port_open = FALSE;		/* record status of Async Port */

void _exit(status)
int status;
{
	union REGS rv;
	if (port_open) {
		asy_clos();			/* close async processing */
	}
	rv.h.ah = 0x4c;		/* call DOS to terminate the program */
	rv.h.al = status;
	intdos(&rv, &rv);
}

void clrscr()
/*  Clear the Screen */
{
	VIDSCROLL(0, 1, 1, 25, 80);	/* clear the whole screen */
	VIDSETCP(1, 1);					/* position to upper left corner */
}

void writeac(a, c)
/* write an attr/char to the current position, then increment the position */
char a, c;
{
	VIDWRTAC(a, c, 1);
	if (curcol < 80) VIDSETCP(currow, curcol+1);
}

void bell()
/* make a bell sound */
{
	VIDWRTTTY('\007');				/* Ctrl-G for a bell sound */
}

void linefeed()
/* perform the line feed function */
{
	if (currow >= 24) VIDSCROLL(1, 1, 1, 24, 80);	/* scroll up 1 line */
	else VIDSETCP(currow + 1, curcol);
}

void cursor_down(n)
/* move cursor down n lines */
int n;
{
	int r;
	r = currow + n;
	if (r > 24) VIDSETCP(24, curcol);
	else VIDSETCP(r, curcol);
}

void erase_eol()
/* erase from cursor to end of line */
{
	VIDWRTAC(ATTR_NORM, ' ', 81-curcol);
}

void erase_bol()
/* erase from beginning of line to cursor */
{
	int oldcol;
	oldcol = curcol;
	VIDSETCP(currow, 1);
	VIDWRTAC(ATTR_NORM, ' ', oldcol);
	VIDSETCP(currow, oldcol);
}

void erase_line()
/* erase entire line */
{
	int oldcol;
	oldcol = curcol;
	VIDSETCP(currow, 1);
	VIDWRTAC(ATTR_NORM, ' ', 80);
	VIDSETCP(currow, oldcol);
}

void erase_eos()
/* erase from cursor to end of screen */
{
	erase_eol();
	if (currow != 24) VIDSCROLL(0, currow + 1, 1, 24, 80);
}

void erase_bos()
/* erase from beginning of screen to cursor */
{
	if (currow != 1) VIDSCROLL(0, 1, 1, currow - 1, 80);
	erase_bol();
}

void erase_scr()
/* erase entire screen */
{
	VIDSCROLL(0, 1, 1, 24, 80);
}
			
void sendstr(s)
/* send a string to the host */
char *s;
{
	char c;
	while ((c = *s++) != 0) asy_send(c);
}

void vt100_kbd(c)
unsigned int c;
/* translate IBM keycodes to VT100 keycodes, then transmit them to the host */
{
	register unsigned int scancode, asciicode;
	scancode = c >> 8;			/* scan code is in high byte */
	asciicode = c & 0x00FF;		/* ascii code is in low byte */
	if (asciicode != 0)			/* check for normal ansi code */
		{
			if (asciicode > 127) bell();		/* VT100 cannot do extended ASCII */
			else asy_send(asciicode);
		}
	else								/* extended keycode */
		{
			switch (scancode)
				{
					case  72:					/* cursor up */
						sendstr("\033[A");
						break;
					case  75:					/* cursor left */
						sendstr("\033[D");
						break;
					case  77:					/* cursor right */
						sendstr("\033[C");
						break;
					case  80:					/* cursor down */
						sendstr("\033[B");
						break;
					case  59:					/* F1 */
						sendstr("\033P");
						break;
					case  60:					/* F2 */
						sendstr("\033Q");
						break;
					case  61:					/* F3 */
						sendstr("\033R");
						break;
					case  62:					/* F4 */
						sendstr("\033S");
						break;
					default: bell();
				}
		}
}

vt100_esc(c, n, p)
/* handle an escape sequence */
char c;						/* Final Character of the sequence */
int n;						/* number of parameters */
int p[];						/* parameter values */
{
	register int i;
	switch (c)
		{
			case 'A':
				if ((n == 0) || (p[1] == 0)) VIDSETCP(currow - 1, curcol);
				else VIDSETCP(currow - p[1], curcol);
				break;
			case 'B':
				if ((n == 0) || (p[1] == 0)) cursor_down(1);
				else cursor_down(p[n]);
				break;
			case 'C':
				if ((n == 0) || (p[1] == 0)) VIDSETCP(currow, curcol + 1);
				else VIDSETCP(currow, curcol + p[n]);
				break;
			case 'D':
				if ((n == 0) || (p[1] == 0)) VIDSETCP(currow, curcol - 1);
				else VIDSETCP(currow, curcol - p[n]);
				break;
			case 'f':
			case 'H':
				if ((n < 2) || (p[2] == 0)) p[2] = 1;
				if ((n < 1) || (p[1] == 0)) p[1] = 1;
				VIDSETCP(p[1], p[2]);
				break;
			case 'J':
				if ((n == 0) || (p[1] == 0)) erase_eos();
				else if (p[1] == 1) erase_bos();
				else if (p[1] == 2) erase_scr();
				break;
			case 'K':
				if ((n == 0) || (p[1] == 0)) erase_eol();
				else if (p[1] == 1) erase_bol();
				else if (p[1] == 2) erase_line();
				break;
			case 'm':
				if (n == 0) curattr = ATTR_NORM;
				else
					{
						for (i=1; i<=n; i++)
							{
								if (p[i] == 0) curattr = ATTR_NORM;
								else curattr = ATTR_BOLD;
							}
					}
				break;
			default: bell();
		}
}

#define MAX_PARM	8		/* max parms in an escape sequence */

void vt100_dsp(c)
char c;
/* Display a character received from the Host */
{
	/* the following variable tells us if we are processing an escape
		sequence, and if so what we expect next.

			0 = not in an escape sequence
			1 = ESC found; next char must be '['
			2 = ESC [ found; waiting for parameter value or the Final Character.
			3 = ESC [ followed by a number found; accumulating values and
					waiting for the Final Character.
	*/
	static int mode = 0;
	static int num_parm, parm[MAX_PARM + 1];
	register int i;

	switch (mode)
		{
			case 0:
				switch (c)
					{
						case 0000:				/* NUL - ignored */
						case 0030:				/* CAN - ignored */
						case 0032:				/* SUB - treated as CAN */
						case 0177:				/* DEL - ignored */
							break;
						case 0007:				/* BEL - sound the bell */
							bell();
							break;
						case 0010:				/* BS - non-destructive backspace */
							VIDSETCP(currow, curcol - 1);
							break;
						case 0011:				/* HT - horizontal tab */
							VIDSETCP(currow, 80);
							break;
						case 0012:				/* LF - line feed */
						case 0013:				/* VT - treated as line feed */
						case 0014:				/* FF - treated as line feed */
							linefeed();
							break;
						case 0015:				/* CR - move to left on cur line */
							VIDSETCP(currow, 1);
							break;
						case 0033:				/* ESC - begin escape sequence */
							mode = 1;
							break;
						default:
							if (c < ' ') bell();			/* bell if not supported */
							else writeac(curattr, c);  /* else display char */
					}
				break;
			case 1:				/* have ESC, need '[' */
				if (c == '[') mode = 2;
				else
					{
						bell();
						mode = 0;
					}
				break;
			case 2:				/* have ESC [, wait for # or Final Char */
				if (isdigit(c))
					{
						num_parm = 1;
						parm[1] = c - '0';
						for (i=2; i<=MAX_PARM; i++) parm[i] = 0;
						mode = 3;
					}
				else if (c == ';')
					{
						num_parm = 2;
						parm[1] = 0;
						parm[2] = 0;
						for (i=3; i<=MAX_PARM; i++) parm[i] = 0;
						mode = 3;
					}
				else
					{
						vt100_esc(c, 0, parm);
						mode = 0;
					}
				break;
			case 3:				/* ESC [ n... */
				if (isdigit(c)) parm[num_parm] = parm[num_parm] * 10 + (c - '0');
				else if (c == ';')
					{
						if (num_parm >= MAX_PARM) bell();
						else num_parm++;
					}
				else
					{
						vt100_esc(c, num_parm, parm);
						mode = 0;
					}
				break;
		}
}
				
void bottom_msg(s)
char *s;
/* Display message on the bottom line */
{
	int oldrow, oldcol;
	char c;

	oldrow = currow;			/* save current cursor position */
	oldcol = curcol;
	VIDSETCP(25, 1);			/* position to bottom line of screen */
	while ((c = *s++) != 0) writeac(ATTR_REV, c); /* display message */
	VIDWRTAC(ATTR_REV, ' ', 81 - curcol);  /* blank out rest of line */
	VIDSETCP(oldrow, oldcol);
}

/*
** Main Program
** Open the communication port, and begin emulating a VT100
*/
main()
{
	register unsigned int c;

	asy_open(COMM_PORT-1, COMM_PARM);	/* Set up for communications */
	port_open = TRUE;						/* remember port is open */
	clrscr();
	bottom_msg("ALT-X to exit");

	for(;;)
		{
			if (kbd_stat())
				{
					c = kbd_getc();
					if (c == 0x2D00) break;		/* Alt-X to exit */
					vt100_kbd(c);
				}
			if (asy_rdy()) vt100_dsp(asy_rcve());
		}

	asy_clos();
	port_open = FALSE;
	VIDSETCP(25, 1);
	VIDWRTAC(ATTR_NORM, ' ', 80);
}
